@@ -20,12 +20,12 @@
  */
 
 #include "consumer.h"
+
 #include "enums.h"
 #include "marshal.h"
+#include "service-monitor.h"
 #include "shell.h"
 
-#include <avahi-client/lookup.h>
-#include <avahi-common/error.h>
 #include <glib/gi18n-lib.h>
 #include <libsoup/soup-session-async.h>
 #include <libsoup/soup-uri.h>
@@ -50,8 +50,8 @@
  *
  *   if (service_name)
  *     consumer = epc_consumer_new_for_name (service_name);
- *   else if (your_app_discover_server (&protocol, &hostname, &port))
- *     consumer = epc_consumer_new (protocol, hostname, port);
+ *   else
+ *     consumer = epc_consumer_new (your_app_find_service ());
  *
  *   value = epc_consumer_lookup (consumer, "glom-settings", NULL, &error);
  *   g_object_unref (consumer);
@@ -83,11 +83,14 @@
  *
  *  if (GTK_RESPONSE_ACCEPT == gtk_dialog_run (GTK_DIALOG (dialog)))
  *   {
- *      const gchar *transport = aui_service_dialog_get_service_type (AUI_SERVICE_DIALOG (dialog));
- *      const gchar *hostname = aui_service_dialog_get_host_name (AUI_SERVICE_DIALOG (dialog));
- *      const gint port = aui_service_dialog_get_port (AUI_SERVICE_DIALOG (dialog));
- * 
- *      consumer = epc_consumer_new (epc_service_type_get_protocol (transport), hostname, port);
+ *      EpcServiceInfo *service =
+ *        epc_service_info_new (aui_service_dialog_get_service_type (AUI_SERVICE_DIALOG (dialog)),
+ *                              aui_service_dialog_get_host_name    (AUI_SERVICE_DIALOG (dialog)),
+ *                              aui_service_dialog_get_port         (AUI_SERVICE_DIALOG (dialog)),
+ *                              aui_service_dialog_get_txt_data     (AUI_SERVICE_DIALOG (dialog)));
+ *
+ *      consumer = epc_consumer_new (service);
+ *      epc_service_info_unref (service);
  *      ...
  *   }
  *  </programlisting>
@@ -116,6 +119,7 @@ enum
   PROP_PROTOCOL,
   PROP_HOSTNAME,
   PROP_PORT,
+  PROP_PATH,
   PROP_USERNAME,
   PROP_PASSWORD
 };
@@ -134,19 +138,29 @@ enum
  */
 struct _EpcConsumerPrivate
 {
-  AvahiClient  *client;
-  GSList       *browsers;
-  SoupSession  *session;
-  GMainLoop    *loop;
+  /* supportive objects */
 
-  gchar        *name;
-  gchar        *domain;
-  gchar        *application;
-  EpcProtocol   protocol;
-  gchar        *hostname;
-  guint16       port;
-  gchar        *username;
-  gchar        *password;
+  EpcServiceMonitor *service_monitor;
+  SoupSession       *session;
+  GMainLoop         *loop;
+
+  /* search parameters */
+
+  gchar       *application;
+  EpcProtocol  protocol;
+
+  /* service credentials */
+
+  gchar       *username;
+  gchar       *password;
+
+  /* service description */
+
+  gchar       *name;
+  gchar       *domain;
+  gchar       *hostname;
+  gchar       *path;
+  guint16      port;
 };
 
 struct _EpcListingState
@@ -157,7 +171,6 @@ struct _EpcListingState
 };
 
 static guint signals[SIGNAL_LAST];
-extern gboolean _epc_debug;
 
 G_DEFINE_TYPE (EpcConsumer, epc_consumer, G_TYPE_OBJECT);
 
@@ -172,7 +185,7 @@ epc_consumer_authenticate_cb (SoupSession  *session G_GNUC_UNUSED,
 {
   EpcConsumer *self = EPC_CONSUMER (data);
 
-  if (G_UNLIKELY (_epc_debug))
+  if (EPC_DEBUG_LEVEL (1))
     g_debug ("%s: path=%s, realm=%s, username=%s, password=%s",
              G_STRLOC, soup_message_get_uri (message)->path,
              auth_realm, *username, *password);
@@ -183,7 +196,7 @@ epc_consumer_authenticate_cb (SoupSession  *session G_GNUC_UNUSED,
   *username = g_strdup (self->priv->username ? self->priv->username : "");
   *password = g_strdup (self->priv->password ? self->priv->password : "");
 
-  if (G_UNLIKELY (_epc_debug))
+  if (EPC_DEBUG_LEVEL (1))
     g_debug ("%s: path=%s, realm=%s, username=%s, password=%s",
              G_STRLOC, soup_message_get_uri (message)->path,
              auth_realm, *username, *password);
@@ -201,17 +214,15 @@ epc_consumer_reauthenticate_cb (SoupSession  *session,
   EpcConsumer *self = EPC_CONSUMER (data);
   gboolean handled = FALSE;
 
-  if (G_UNLIKELY (_epc_debug))
+  if (EPC_DEBUG_LEVEL (1))
     g_debug ("%s: path=%s, realm=%s, username=%s, password=%s, handled=%d",
              G_STRLOC, soup_message_get_uri (message)->path,
              auth_realm, *username, *password, handled);
 
-  epc_shell_enter ();
-  g_signal_emit (self, signals[SIGNAL_AUTHENTICATE], 
+  g_signal_emit (self, signals[SIGNAL_AUTHENTICATE],
                  0, auth_realm, &handled);
-  epc_shell_leave ();
 
-  if (G_UNLIKELY (_epc_debug))
+  if (EPC_DEBUG_LEVEL (1))
     g_debug ("%s: path=%s, realm=%s, username=%s, password=%s, handled=%d",
              G_STRLOC, soup_message_get_uri (message)->path,
              auth_realm, *username, *password, handled);
@@ -224,8 +235,6 @@ epc_consumer_reauthenticate_cb (SoupSession  *session,
 static void
 epc_consumer_init (EpcConsumer *self)
 {
-  epc_shell_ref ();
-
   self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self, EPC_TYPE_CONSUMER, EpcConsumerPrivate);
   self->priv->loop = g_main_loop_new (NULL, FALSE);
   self->priv->session = soup_session_async_new ();
@@ -262,7 +271,7 @@ epc_consumer_set_property (GObject      *object,
         break;
 
       case PROP_PROTOCOL:
-        g_return_if_fail (NULL == self->priv->browsers &&
+        g_return_if_fail (NULL == self->priv->service_monitor &&
                           NULL == self->priv->hostname);
         self->priv->protocol = g_value_get_enum (value);
         break;
@@ -277,6 +286,11 @@ epc_consumer_set_property (GObject      *object,
         self->priv->port = g_value_get_int (value);
         break;
 
+      case PROP_PATH:
+        g_assert (NULL == self->priv->path);
+        self->priv->path = g_value_dup_string (value);
+        break;
+
       case PROP_USERNAME:
         g_free (self->priv->username);
         self->priv->username = g_value_dup_string (value);
@@ -327,6 +341,10 @@ epc_consumer_get_property (GObject    *object,
         g_value_set_int (value, self->priv->port);
         break;
 
+      case PROP_PATH:
+        g_value_set_string (value, self->priv->path);
+        break;
+
       case PROP_USERNAME:
         g_value_set_string (value, self->priv->username);
         break;
@@ -342,130 +360,44 @@ epc_consumer_get_property (GObject    *object,
 }
 
 static void
-epc_consumer_service_resolver_cb (AvahiServiceResolver   *resolver,
-                                  AvahiIfIndex            interface G_GNUC_UNUSED,
-                                  AvahiProtocol           protocol G_GNUC_UNUSED,
-                                  AvahiResolverEvent      event G_GNUC_UNUSED,
-                                  const char             *name G_GNUC_UNUSED,
-                                  const char             *type,
-                                  const char             *domain G_GNUC_UNUSED,
-                                  const char             *hostname,
-                                  const AvahiAddress     *a G_GNUC_UNUSED,
-                                  uint16_t                port,
-                                  AvahiStringList        *txt G_GNUC_UNUSED,
-                                  AvahiLookupResultFlags  flags G_GNUC_UNUSED,
-                                  void                   *data G_GNUC_UNUSED)
+epc_consumer_service_found_cb (EpcConsumer    *self,
+                               const gchar    *name,
+                               EpcServiceInfo *info)
 {
-  EpcConsumer *self = data;
-  EpcProtocol transport;
+  const gchar *type = epc_service_info_get_service_type (info);
+  EpcProtocol transport = epc_service_type_get_protocol (type);
 
-  if (G_UNLIKELY (_epc_debug))
-    g_debug ("%s: Service resolved: type='%s', hostname='%s', port=%d", 
-             G_STRLOC, type, hostname, port);
+  const gchar *path = epc_service_info_get_detail (info, "path");
+  const gchar *host = epc_service_info_get_host (info);
+  guint port = epc_service_info_get_port (info);
+
+  if (EPC_DEBUG_LEVEL (1))
+    g_debug ("%s: Service resolved: type='%s', host='%s', port=%d, path='%s'", 
+             G_STRLOC, type, host, port, path);
+
+  if (name && strcmp (name, self->priv->name))
+    return;
 
-  g_assert (EPC_PROTOCOL_HTTP > EPC_PROTOCOL_UNKNOWN);
   g_assert (EPC_PROTOCOL_HTTPS > EPC_PROTOCOL_HTTP);
 
-  transport = epc_service_type_get_protocol (type);
-
   if (transport > self->priv->protocol)
     {
-      if (G_UNLIKELY (_epc_debug))
-        g_debug ("%s: Upgrading to '%s' protocol", G_STRLOC, 
-                 epc_protocol_get_service_type (transport));
-
-      g_signal_emit (self, signals[SIGNAL_PUBLISHER_RESOLVED],
-                     0, transport, hostname, (guint)port);
+      if (EPC_DEBUG_LEVEL (1))
+        g_debug ("%s: Upgrading to %s protocol", G_STRLOC, epc_protocol_get_service_type (transport));
 
+      g_signal_emit (self, signals[SIGNAL_PUBLISHER_RESOLVED], 0, transport, host, port);
       self->priv->protocol = transport;
     }
 
   g_main_loop_quit (self->priv->loop);
 
+  g_free (self->priv->path);
   g_free (self->priv->hostname);
-  self->priv->hostname = g_strdup (hostname);
+
+  /* Use /get path as fallback for libepc-0.2 publishers */
+  self->priv->path = g_strdup (path ? path : "/get");
+  self->priv->hostname = g_strdup (host);
   self->priv->port = port;
-
-  avahi_service_resolver_free (resolver);
-}
-
-static void
-epc_consumer_service_brower_cb (AvahiServiceBrowser    *browser,
-                                AvahiIfIndex            interface,
-                                AvahiProtocol           protocol,
-                                AvahiBrowserEvent       event,
-                                const char             *name,
-                                const char             *type,
-                                const char             *domain,
-                                AvahiLookupResultFlags  flags G_GNUC_UNUSED,
-                                void                   *data)
-{
-  EpcConsumer *self = data;
-  gint error;
-
-  if (G_UNLIKELY (_epc_debug))
-    g_debug ("%s: event=%d, name=`%s', type=`%s', domain=`%s'", 
-             G_STRLOC, event, name, type, domain);
-
-  switch (event)
-    {
-      case AVAHI_BROWSER_REMOVE:
-        break;
-
-      case AVAHI_BROWSER_FAILURE:
-        error = avahi_client_errno (avahi_service_browser_get_client (browser));
-        g_warning ("%s: %s (%d)", G_STRFUNC, avahi_strerror (error), error);
-        break;
-
-      default:
-        if (name && g_str_equal (name, self->priv->name))
-          avahi_service_resolver_new (self->priv->client, interface,
-                                      protocol, name, type, domain, AVAHI_PROTO_UNSPEC,
-                                      0, epc_consumer_service_resolver_cb, self);
-
-        break;
-    }
-}
-
-static void
-epc_consumer_create_service_browsers (EpcConsumer *self,
-                                      EpcProtocol  protocol,
-                                                   ...)
-{
-  AvahiServiceBrowser *browser;
-  va_list args;
-
-  va_start (args, protocol);
-
-  while (protocol > EPC_PROTOCOL_UNKNOWN)
-    {
-      gchar *service_type;
-
-      service_type = epc_service_type_new (protocol, self->priv->application);
-      protocol = va_arg (args, EpcProtocol);
-
-      if (NULL == service_type)
-        continue;
-
-      browser = avahi_service_browser_new (self->priv->client,
-                                           AVAHI_IF_UNSPEC, AVAHI_PROTO_UNSPEC,
-                                           service_type, self->priv->domain,
-                                           0, epc_consumer_service_brower_cb, self);
-
-      if (NULL == browser)
-        {
-          gint error = avahi_client_errno (self->priv->client);
-
-          g_warning ("%s: Cannot create browser for service type '%s': %s (%d)",
-                     G_STRFUNC, service_type, avahi_strerror (error), error);
-
-          continue;
-      }
-
-      self->priv->browsers = g_slist_prepend (self->priv->browsers, browser);
-    }
-
-  va_end (args);
 }
 
 static void
@@ -478,22 +410,14 @@ epc_consumer_constructed (GObject *object)
 
   if (!self->priv->hostname)
     {
-      GError *error = NULL;
+      self->priv->service_monitor = epc_service_monitor_new (self->priv->application,
+                                                             self->priv->domain,
+                                                             self->priv->protocol,
+                                                             EPC_PROTOCOL_UNKNOWN);
 
-      self->priv->client = epc_shell_create_avahi_client (AVAHI_CLIENT_NO_FAIL,
-                                                          NULL, NULL, &error);
-
-      if (NULL == self->priv->client)
-        {
-          g_warning ("%s: %s", G_STRFUNC, error->message);
-          g_error_free (error);
-          return;
-        }
-
-      if (EPC_PROTOCOL_UNKNOWN != self->priv->protocol)
-        epc_consumer_create_service_browsers (self, self->priv->protocol, 0);
-      else
-        epc_consumer_create_service_browsers (self, EPC_PROTOCOL_HTTP, EPC_PROTOCOL_HTTPS, 0);
+      g_signal_connect_swapped (self->priv->service_monitor, "service-found",
+                                G_CALLBACK (epc_consumer_service_found_cb),
+                                self);
     }
 }
 
@@ -502,16 +426,10 @@ epc_consumer_dispose (GObject *object)
 {
   EpcConsumer *self = EPC_CONSUMER (object);
 
-  while (self->priv->browsers)
+  if (self->priv->service_monitor)
     {
-      avahi_service_browser_free (self->priv->browsers->data);
-      self->priv->browsers = g_slist_delete_link (self->priv->browsers, self->priv->browsers);
-    }
-
-  if (self->priv->client)
-    {
-      avahi_client_free (self->priv->client);
-      self->priv->client = NULL;
+      g_object_unref (self->priv->service_monitor);
+      self->priv->service_monitor = NULL;
     }
 
   if (self->priv->session)
@@ -544,14 +462,10 @@ epc_consumer_dispose (GObject *object)
   g_free (self->priv->password);
   self->priv->password = NULL;
 
-  G_OBJECT_CLASS (epc_consumer_parent_class)->dispose (object);
-}
+  g_free (self->priv->path);
+  self->priv->path = NULL;
 
-static void
-epc_consumer_finalize (GObject *object)
-{
-  G_OBJECT_CLASS (epc_consumer_parent_class)->finalize (object);
-  epc_shell_unref ();
+  G_OBJECT_CLASS (epc_consumer_parent_class)->dispose (object);
 }
 
 static void
@@ -562,7 +476,6 @@ epc_consumer_class_init (EpcConsumerClass *cls)
   oclass->set_property = epc_consumer_set_property;
   oclass->get_property = epc_consumer_get_property;
   oclass->constructed = epc_consumer_constructed;
-  oclass->finalize = epc_consumer_finalize;
   oclass->dispose = epc_consumer_dispose;
 
   g_object_class_install_property (oclass, PROP_NAME,
@@ -609,6 +522,14 @@ epc_consumer_class_init (EpcConsumerClass *cls)
                                                      G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
                                                      G_PARAM_STATIC_BLURB));
 
+  g_object_class_install_property (oclass, PROP_PATH,
+                                   g_param_spec_string ("path", "Path",
+                                                        "The path the publisher uses for contents",
+                                                        "/contents",
+                                                        G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY |
+                                                        G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
+                                                        G_PARAM_STATIC_BLURB));
+
   g_object_class_install_property (oclass, PROP_USERNAME,
                                    g_param_spec_string ("username", "User Name",
                                                         "The user name to use for authentication",
@@ -630,11 +551,11 @@ epc_consumer_class_init (EpcConsumerClass *cls)
    * @consumer: the #EpcConsumer emitting the signal
    * @realm: the realm being authenticated to
    *
-   * Emitted when the #EpcConsumer requires authentication. The signal 
-   * handler should provide these credentials, which may come from the 
+   * Emitted when the #EpcConsumer requires authentication. The signal
+   * handler should provide these credentials, which may come from the
    * user or from cached information by setting the #EpcConsumer:username
    * and #EpcConsumer:password properties. When providing credentials
-   * the signal handler also has to return %TRUE to stop signal emission.
+   * the signal handler should return %TRUE to stop signal emission.
    *
    * If the provided credentials fail then the signal will be emmitted again.
    *
@@ -676,32 +597,37 @@ epc_consumer_class_init (EpcConsumerClass *cls)
 
 /**
  * epc_consumer_new:
- * @protocol: the transport protocol to use
- * @hostname: the publisher's host name
- * @port:  the publisher's TCP/IP port
+ * @service: the publisher's service description
  *
  * Creates a new #EpcConsumer object and associates it with a known
- * #EpcPublisher. Values for @protocol, @hostname and @port can be retrieved, 
- * for instance, by using the service selection dialog of avahi-ui
- * (#AuiServiceDialog). Call #epc_service_type_get_protocol to convert
- * the service-type provided by that dialog to an #EpcProtocol value.
+ * #EpcPublisher. The @service description can be retrieved, for instance,
+ * by using #EpcServiceMonitor, or by using the service selection dialog
+ * of <citetitle>avahi-ui</citetitle> (#AuiServiceDialog).
  *
- * The connection is not established until #epc_consumer_lookup
- * is called to retrieve values.
+ * The connection is not established until functions like #epc_consumer_lookup,
+ * #epc_consumer_list or #epc_consumer_resolve_publisher are called.
  *
  * Returns: The newly created #EpcConsumer object
  */
 EpcConsumer*
-epc_consumer_new (EpcProtocol  protocol,
-                  const gchar *hostname,
-                  guint16      port)
+epc_consumer_new (const EpcServiceInfo *service)
 {
-  g_return_val_if_fail (EPC_PROTOCOL_UNKNOWN != protocol, NULL);
-  g_return_val_if_fail (NULL != hostname, NULL);
-  g_return_val_if_fail (port > 0, NULL);
+  EpcProtocol protocol;
+  const gchar *type;
 
-  return g_object_new (EPC_TYPE_CONSUMER, "protocol", protocol,
-                       "hostname", hostname, "port", port, NULL);
+  g_return_val_if_fail (EPC_IS_SERVICE_INFO (service), NULL);
+
+  type = epc_service_info_get_service_type (service);
+  protocol = epc_service_type_get_protocol (type);
+
+  g_return_val_if_fail (EPC_PROTOCOL_UNKNOWN != protocol, NULL);
+
+  return g_object_new (EPC_TYPE_CONSUMER,
+                       "protocol", protocol,
+                       "hostname", epc_service_info_get_host (service),
+                       "port", epc_service_info_get_port (service),
+                       "path", epc_service_info_get_detail (service, "path"),
+                       NULL);
 }
 
 /**
@@ -743,7 +669,7 @@ epc_consumer_new_for_name (const gchar *name)
  *  in the past when you added it to your list of recently used publishers,
  *  but you do not know if it still exists.
  *
- *  To let your users choose from an up-to-date service list, you have to 
+ *  To let your users choose from an up-to-date service list, you have to
  *  use a dynamic service list as provided by avahi-ui for choosing a service
  *  and pass the information this widget provides (hostname, port, protocol)
  *  to #epc_consumer_new.
@@ -933,32 +859,28 @@ epc_consumer_create_request (EpcConsumer *self,
                              const gchar *path)
 {
   SoupMessage *request = NULL;
+  char *request_uri;
 
   if (NULL == path)
     path = "/";
 
   g_assert ('/' == path[0]);
 
-  if (epc_consumer_resolve_publisher (self, EPC_CONSUMER_DEFAULT_TIMEOUT))
-    {
-      char *request_uri;
+  g_return_val_if_fail (NULL != self->priv->hostname, NULL);
+  g_return_val_if_fail (self->priv->port > 0, NULL);
 
-      g_return_val_if_fail (NULL != self->priv->hostname, NULL);
-      g_return_val_if_fail (self->priv->port > 0, NULL);
+  request_uri = epc_protocol_build_uri (self->priv->protocol,
+                                        self->priv->hostname,
+                                        self->priv->port,
+                                        path);
 
-      request_uri = epc_protocol_build_uri (self->priv->protocol,
-                                            self->priv->hostname,
-                                            self->priv->port,
-                                            path);
+  g_return_val_if_fail (NULL != request_uri, NULL);
 
-      g_return_val_if_fail (NULL != request_uri, NULL);
+  if (EPC_DEBUG_LEVEL (1))
+    g_debug ("%s: Connecting to `%s'", G_STRLOC, request_uri);
 
-      if (G_UNLIKELY (_epc_debug))
-        g_debug ("%s: Connecting to `%s'", G_STRLOC, request_uri);
-
-      request = soup_message_new ("GET", request_uri);
-      g_free (request_uri);
-    }
+  request = soup_message_new ("GET", request_uri);
+  g_free (request_uri);
 
   return request;
 }
@@ -993,6 +915,11 @@ epc_consumer_set_http_error (GError     **error,
  * #EPC_HTTP_ERROR. Error codes are taken from the #SoupKnownStatusCode
  * enumeration.
  *
+ * For instance, the error code will be #SOUP_STATUS_FORBIDDEN if
+ * authentication failed (see #epc_publisher_set_auth_handler). You must
+ * include <filename class="headerfile">libsoup/soup-status.h</filename>
+ * to use this error code.
+ *
  * The returned buffer should be freed when no longer needed.
  *
  * See the description of #EpcPublisher for discussion of %NULL values.
@@ -1000,7 +927,7 @@ epc_consumer_set_http_error (GError     **error,
  * Returns: A copy of the publisher's value for the the requested @key,
  * or %NULL when an error occurred.
  */
-gchar*
+gpointer
 epc_consumer_lookup (EpcConsumer  *self,
                      const gchar  *key,
                      gsize        *length,
@@ -1008,22 +935,26 @@ epc_consumer_lookup (EpcConsumer  *self,
 {
   SoupMessage *request = NULL;
   gchar *contents = NULL;
-  gchar *path = NULL;
   gint status = 0;
 
   g_return_val_if_fail (EPC_IS_CONSUMER (self), NULL);
   g_return_val_if_fail (NULL != key, NULL);
 
-  path = g_strconcat ("/get/", key, NULL);
-  request = epc_consumer_create_request (self, path);
-  g_free (path);
+  if (epc_consumer_resolve_publisher (self, EPC_CONSUMER_DEFAULT_TIMEOUT))
+    {
+      gchar *keyuri = NULL;
+      gchar *path = NULL;
+
+      keyuri = soup_uri_encode (key, NULL);
+      path = g_strconcat (self->priv->path, "/", keyuri, NULL);
+      request = epc_consumer_create_request (self, path);
+
+      g_free (keyuri);
+      g_free (path);
+    }
 
   if (request)
-    {
-      epc_shell_leave ();
-      status = soup_session_send_message (self->priv->session, request);
-      epc_shell_enter ();
-    }
+    status = soup_session_send_message (self->priv->session, request);
   else
     status = SOUP_STATUS_CANT_RESOLVE;
 
@@ -1170,6 +1101,8 @@ epc_consumer_list_parser_text (GMarkupParseContext *context G_GNUC_UNUSED,
  *  g_list_free (keys);
  * </programlisting>
  *
+ * See also #epc_publisher_list for creating custom listings.
+ *
  * Returns: A newly allocated list of keys, or %NULL when an error occurred.
  */
 GList*
@@ -1177,24 +1110,22 @@ epc_consumer_list (EpcConsumer  *self,
                    const gchar  *pattern G_GNUC_UNUSED,
                    GError      **error G_GNUC_UNUSED)
 {
-  EpcListingState state;
   SoupMessage *request = NULL;
-  gchar *path = NULL;
+  EpcListingState state;
   gint status = 0;
 
   g_return_val_if_fail (EPC_IS_CONSUMER (self), NULL);
   g_return_val_if_fail (NULL == pattern || *pattern, NULL);
 
-  path = g_strconcat ("/list/", pattern, NULL);
-  request = epc_consumer_create_request (self, path);
-  g_free (path);
+  if (epc_consumer_resolve_publisher (self, EPC_CONSUMER_DEFAULT_TIMEOUT))
+    {
+      gchar *path = g_strconcat ("/list/", pattern, NULL);
+      request = epc_consumer_create_request (self, path);
+      g_free (path);
+    }
 
   if (request)
-    {
-      epc_shell_leave ();
-      status = soup_session_send_message (self->priv->session, request);
-      epc_shell_enter ();
-    }
+    status = soup_session_send_message (self->priv->session, request);
   else
     status = SOUP_STATUS_CANT_RESOLVE;
 
@@ -1233,5 +1164,3 @@ epc_http_error_quark (void)
 {
   return g_quark_from_static_string ("epc-http-error-quark");
 }
-
-/* vim: set sw=2 sta et spl=en spell: */
